INT = 'int'
FLOAT = 'float'
BOOL = 'bool'
STR = 'str'
DICT = 'dict'
DICT_OF_DICTS = 'dict_of_dicts'
URI = 'uri'
AUTH = 'auth'
LIST = 'list'
GCS_PATH = 'gcs_path'
S3_PATH = 's3_path'
AZURE_PATH = 'azure_path'
PATH = 'path'
METRIC = 'metric'
METADATA = 'metadata'

VALUES = {
    INT,
    FLOAT,
    BOOL,
    STR,
    DICT,
    DICT_OF_DICTS,
    URI,
    AUTH,
    LIST,
    GCS_PATH,
    S3_PATH,
    AZURE_PATH,
    PATH,
    METRIC,
    METADATA
}
